body {
  padding: 0 !important;
}

header.header-top.fixed-top {
  display: none;
}

div.sidenav {
  top: 0;
}

.toc-cards > ul {
  align-content: stretch;
  align-items: stretch;
  display: flex;
  flex-flow: row wrap;
  gap: 10px 20px;
  justify-content: space-between;
  padding: 0;
}

.toc-cards > ul > li {
  background: white;
  border: 2px solid #a8acaf;
  border-radius: 10px;
  list-style-type: none;
}

.toc-cards > ul > li > a {
  align-items: center;
  color: #2c2c2c;
  display: flex;
  font-family: "Gilroy", sans-serif;
  height: 100%;
  justify-content: center;
  text-align: center;
  text-decoration: none;
}

.toc-cards > ul > li > a:hover {
  color: #6b6b6b;
}

.toc-cards.toc-cards-narrow > ul > li {
  font-size: 150%;
  min-height: 150px;
  width: calc(50% - 20px);
}
